import hmac
import hashlib
import json
import logging

import pandas
import pandas as pd
import requests
from urllib.parse import urlencode

from onecall.base import utils
from onecall.base.exchange import Exchange
from onecall.base import urls


class FTX(Exchange):
    """
    FTX API Class
    """

    # Intervals
    INTERVAL_15 = 15
    INTERVAL_60 = 60
    INTERVAL_300 = 300
    INTERVAL_900 = 900
    INTERVAL_3600 = 3600
    INTERVAL_14400 = 14400
    INTERVAL_86400 = 86400

    # Constants for Order side
    BUY_SIDE = 'buy'
    SELL_SIDE = 'sell'

    def __init__(self, key=None, secret=None, **kwargs):
        """
        FTX API class
        https://docs.ftx.com/#rest-api

        :param key: API key
        :param secret: Secret key
        """
        self._path_config = {
            "get_positions": {"method": "GET", "path": "/positions", "rate_limit": 50},
            "cancel_orders": {"method": "DELETE", "path": "/orders", "rate_limit": 50},
            "get_data": {"method": "GET", "path": "/markets/{symbol}/candles", "rate_limit": 50},
            "get_orderbook": {"method": "GET", "path": "/markets/{symbol}/orderbook", "rate_limit": 50},
            "get_balance": {"method": "GET", "path": "/wallet/balances", "rate_limit": 50},
            "market_order": {"method": "POST", "path": "/orders", "rate_limit": 50},
            "limit_order": {"method": "POST", "path": "/orders", "rate_limit": 50},
            "get_closed_orders": {"method": "GET", "path": "/orders/history", "rate_limit": 50},
            "get_open_orders": {"method": "GET", "path": "/orders", "rate_limit": 50}
        }
        if not kwargs.get("base_url"):
            kwargs["base_url"] = urls.FTX_FUT_BASE_URL
        super().__init__(key, secret, **kwargs)
        return

    def get_positions(self, symbol):
        """
        API to get current positions
        https://docs.ftx.com/#get-positions

        :param symbol: symbol
        :return: {
              "success": true,
              "result": [
                {
                  "cost": -31.7906,
                  "cumulativeBuySize": 1.2,
                  "cumulativeSellSize": 0.0,
                  "entryPrice": 138.22,
                  "estimatedLiquidationPrice": 152.1,
                  "future": "ETH-PERP",
                  "initialMarginRequirement": 0.1,
                  "longOrderSize": 1744.55,
                  "maintenanceMarginRequirement": 0.04,
                  "netSize": -0.23,
                  "openSize": 1744.32,
                  "realizedPnl": 3.39441714,
                  "recentAverageOpenPrice": 135.31,
                  "recentBreakEvenPrice": 135.31,
                  "recentPnl": 3.1134,
                  "shortOrderSize": 1732.09,
                  "side": "sell",
                  "size": 0.23,
                  "unrealizedPnl": 0,
                  "collateralUsed": 3.17906
                }
              ]
            }
        """
        response = self._signed_request(self._path_config.get("get_positions").get("method"),
                                        self._path_config.get("get_positions").get("path"))
        if response.get("result"):
            return list(filter(lambda pos: pos.get("future") == symbol, response.get("result")))
        return response

    def cancel_orders(self, symbol):
        """
        API to cancel all orders
        https://docs.ftx.com/#cancel-all-orders

        :param symbol: contract symbol
        :return: {
                  "success": true,
                  "result": "Orders queued for cancelation"
                }
        """
        payload = {
          "market": symbol,
        }
        response = self._signed_request(self._path_config.get("cancel_orders").get("method"),
                                        self._path_config.get("cancel_orders").get("path"),
                                        payload)
        return response

    def get_data(self, symbol, interval=15, **kwargs):
        """
        API to get data
        https://docs.ftx.com/#get-historical-prices

        :param symbol: contract symbol
        :param interval: interval
        :keyword start_time: start time
        :keyword end_time: end time
        :keyword is_dataframe: convert to dataframe or not
        :return: {
              "success": true,
              "result": [
                {
                  "close": 11055.25,
                  "high": 11089.0,
                  "low": 11043.5,
                  "open": 11059.25,
                  "startTime": "2019-06-24T17:15:00+00:00",
                  "volume": 464193.95725
                }
              ]
            }
        """
        params = {
            "resolution": interval,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("get_data").get("method"),
                                        self._path_config.get("get_data").get("path").format(symbol=symbol),
                                        params)
        if kwargs.get("is_dataframe") and response.get("result"):
            try:
                return pd.DataFrame(response.get("result"))
            except Exception as e:
                logging.error("failed to create dataframe: ", e)
        return response

    def get_orderbook(self, symbol, limit=100, is_dataframe=False):
        """
        API to get orderbook
        https://docs.ftx.com/#get-orderbook

        :param symbol: contract symbol
        :param limit: data limit
        :param is_dataframe: convert the data to dataframe or not
        :return: {
              "success": true,
              "result": {
                "asks": [
                  [
                    4114.25,
                    6.263
                  ]
                ],
                "bids": [
                  [
                    4112.25,
                    49.29
                  ]
                ]
              }
            }
        """
        params = {
            "depth": limit
        }
        response = self._signed_request(self._path_config.get("get_orderbook").get("method"),
                                        self._path_config.get("get_orderbook").get("path").format(symbol=symbol),
                                        params)
        if is_dataframe and response.get("result"):
            try:
                columns = ['price', 'QTY']
                ask = pd.DataFrame(response["result"]["bids"], columns=columns)
                ask["type"] = ["ask" for i in range(0, ask.shape[0])]
                bid = pd.DataFrame(response["result"]["asks"], columns=columns)
                bid["type"] = ["bid" for i in range(0, bid.shape[0])]
                ask = pd.concat([bid, ask], ignore_index=True)
                return ask
            except Exception as e:
                logging.error("failed to create dataframe: ", e)
        return response

    def get_balance(self):
        """
        API to get the balance
        https://docs.ftx.com/#get-balances

        :return: {
              "success": true,
              "result": {
                "main": [
                  {
                    "coin": "USDTBEAR",
                    "free": 2320.2,
                    "spotBorrow": 0.0,
                    "total": 2340.2,
                    "usdValue": 2340.2,
                    "availableWithoutBorrow": 2320.2
                  },
                  {
                    "coin": "BTC",
                    "free": 2.0,
                    "spotBorrow": 0.0,
                    "total": 3.2,
                    "usdValue": 23456.7,
                    "availableWithoutBorrow": 2.0
                  }
                ],
            }
        }
        """
        response = self._signed_request(self._path_config.get("get_balance").get("method"),
                                        self._path_config.get("get_balance").get("path"))
        return response

    def market_order(self, symbol, side, quantity, **kwargs):
        """
        API to place market order
        https://docs.ftx.com/#place-order

        :param symbol: contract symbol
        :param side: buy/sell
        :param quantity: order quantity
        :return: {
              "success": true,
              "result": {
                "createdAt": "2019-03-05T09:56:55.728933+00:00",
                "filledSize": 0,
                "future": "XRP-PERP",
                "id": 9596912,
                "market": "XRP-PERP",
                "price": 0.306525,
                "remainingSize": 31431,
                "side": "sell",
                "size": 31431,
                "status": "open",
                "type": "market",
                "reduceOnly": false,
                "ioc": false,
                "postOnly": false,
                "clientId": null,
              }
            }
        """
        payload = {
            "market": symbol,
            "side": side,
            "type": "market",
            "size": quantity,
            "price": None,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("market_order").get("method"),
                                        self._path_config.get("market_order").get("path"),
                                        data=payload)
        return response

    def limit_order(self, symbol, side, quantity, price, **kwargs):
        """
        API to place limit order
        https://docs.ftx.com/#place-order

        :param symbol: contract symbol
        :param side: buy/sell
        :param quantity: order quantity
        :param price: order price
        :keyword post_only: for post only request
        :return: {
              "success": true,
              "result": {
                "createdAt": "2019-03-05T09:56:55.728933+00:00",
                "filledSize": 0,
                "future": "XRP-PERP",
                "id": 9596912,
                "market": "XRP-PERP",
                "price": 0.306525,
                "remainingSize": 31431,
                "side": "sell",
                "size": 31431,
                "status": "open",
                "type": "limit",
                "reduceOnly": false,
                "ioc": false,
                "postOnly": false,
                "clientId": null,
              }
            }
        """
        payload = {
            "market": symbol,
            "side": side,
            "price": price,
            "type": "limit",
            "size": quantity,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("limit_order").get("method"),
                                        self._path_config.get("limit_order").get("path"),
                                        data=payload)
        return response

    def get_closed_orders(self):
        """
        API to get all closed orders
        https://docs.ftx.com/#get-order-history

        :return: {
              "success": true,
              "result": [
                {
                  "avgFillPrice": 10135.25,
                  "clientId": null,
                  "createdAt": "2019-06-27T15:24:03.101197+00:00",
                  "filledSize": 0.001,
                  "future": "BTC-PERP",
                  "id": 257132591,
                  "ioc": false,
                  "market": "BTC-PERP",
                  "postOnly": false,
                  "price": 10135.25,
                  "reduceOnly": false,
                  "remainingSize": 0.0,
                  "side": "buy",
                  "size": 0.001,
                  "status": "closed",
                  "type": "limit"
                },
              ],
              "hasMoreData": false,
            }
        """
        response = self._signed_request(self._path_config.get("get_closed_orders").get("method"),
                                        self._path_config.get("get_closed_orders").get("path"))
        if response.get("result"):
            return list(filter(lambda order: order.get("status", "") == "closed", response.get("result")))
        return response

    def get_open_orders(self):
        """
        API to get all open orders
        https://docs.ftx.com/#get-open-orders

        :return: {
              "success": true,
              "result": [
                {
                  "createdAt": "2019-03-05T09:56:55.728933+00:00",
                  "filledSize": 10,
                  "future": "XRP-PERP",
                  "id": 9596912,
                  "market": "XRP-PERP",
                  "price": 0.306525,
                  "avgFillPrice": 0.306526,
                  "remainingSize": 31421,
                  "side": "sell",
                  "size": 31431,
                  "status": "open",
                  "type": "limit",
                  "reduceOnly": false,
                  "ioc": false,
                  "postOnly": false,
                  "clientId": null
                }
              ]
            }
        """
        response = self._signed_request(self._path_config.get("get_open_orders").get("method"),
                                        self._path_config.get("get_open_orders").get("path"))
        return response

    def _signed_request(self, method, url, params=None, data=None):
        ts = str(utils.get_current_timestamp())
        params = urlencode(params) if params else ""
        data = json.dumps(data) if data else None
        headers = self._get_request_credentials(ts, method, url, params, data)
        response = self.send_request(method, url, headers, params, data)
        return response

    def _get_sign(self, data):
        m = hmac.new(self.secret.encode(), data.encode(), hashlib.sha256)
        return m.hexdigest()

    def _get_request_credentials(self, ts, method, path, params="", data=None):
        url = f'{self.base_url}{path}?{params}'
        request = requests.Request(method=method, url=url, data=data)
        prepared = request.prepare()
        signature_payload = f'{ts}{prepared.method}{prepared.path_url}'
        if prepared.body:
            signature_payload += prepared.body
        sing = self._get_sign(signature_payload)
        if self.base_url == urls.FTXUS_FUT_BASE_URL:
            header = {
                "FTXUS-KEY": self.key,
                "FTXUS-SIGN": sing,
                "FTXUS-TS": ts
            }
        else:
            header = {
                "FTX-KEY": self.key,
                "FTX-SIGN": sing,
                "FTX-TS": ts
            }
        return header



class ServerException(Exception):
    def __init__(self, message):
        super().__init__(message)


class ClientException(Exception):
    def __init__(self, message):
        super().__init__(message)


class PandasDataframeException(Exception):
    def __init__(self, message):
        super().__init__(message)



import datetime
import hmac
import hashlib
import logging
import pandas as pd
from urllib.parse import urlencode

from onecall.base import utils
from onecall.base.exchange import Exchange
from onecall.base import urls


class BinanceSpot(Exchange):
    """
    Binance API class
    https://binance-docs.github.io/apidocs/spot/en/#general-api-information
    """
    def __init__(self, key=None, secret=None, debug=False, **kwargs):
        """
        Binance API

        :param key: api key
        :param secret: secret key
        :param debug: flag to switch to testnet
        """
        self._path_config = {
            "cancel_orders": {"method": "DELETE", "path": "/api/v3/openOrders", "rate_limit": 50},
            "get_data": {"method": "GET", "path": "/api/v3/klines", "rate_limit": 50},
            "get_orderbook": {"method": "GET", "path": "/api/v3/depth", "rate_limit": 50},
            "get_balance": {"method": "GET", "path": "/api/v3/account", "rate_limit": 50},
            "market_order": {"method": "POST", "path": "/api/v3/order", "rate_limit": 50},
            "limit_order": {"method": "POST", "path": "/api/v3/order", "rate_limit": 50},
            "get_closed_orders": {"method": "GET", "path": "/api/v3/allOrders", "rate_limit": 50},
            "get_open_orders": {"method": "GET", "path": "/api/v3/openOrders", "rate_limit": 50}
        }
        self._LIMIT = 500

        # Constants for Order side
        self.BUY_SIDE = 'BUY'
        self.SELL_SIDE = 'SELL'

        # binance interval
        self.INTERVAL_1m = '1m'
        self.INTERVAL_3m = '3m'
        self.INTERVAL_5m = '5m'
        self.INTERVAL_15m = '15m'
        self.INTERVAL_30m = '30m'
        self.INTERVAL_1H = '1h'
        self.INTERVAL_2H = '2h'
        self.INTERVAL_4H = '4h'
        self.INTERVAL_6H = '6h'
        self.INTERVAL_8H = '8h'
        self.INTERVAL_12H = '12h'
        self.INTERVAL_1D = '1d'
        self.INTERVAL_3D = '3d'
        self.INTERVAL_1W = '1w'
        self.INTERVAL_1M = '1M'

        if not debug:
            kwargs["base_url"] = urls.BINANCE_SPOT_BASE_URL
        else:
            kwargs["base_url"] = urls.BINANCE_SPOT_TEST_BASE_URL
        super().__init__(key, secret, **kwargs)
        return

    def cancel_orders(self, symbol: str):
        """
        API to cancel all the active orders
        https://binance-docs.github.io/apidocs/spot/en/#cancel-all-open-orders-on-a-symbol-trade

        :param symbol: future symbol
        :return: [
            {
                "clientOrderId": "myOrder1",
                "cumQty": "0",
                "cumQuote": "0",
                "executedQty": "0",
                "orderId": 283194212,
                "origQty": "11",
                "origType": "TRAILING_STOP_MARKET",
                "price": "0",
                "reduceOnly": false,
                "side": "BUY",
                "positionSide": "SHORT",
                "status": "CANCELED",
                "stopPrice": "9300",                // please ignore when order type is TRAILING_STOP_MARKET
                "closePosition": false,   // if Close-All
                "symbol": "BTCUSDT",
                "timeInForce": "GTC",
                "type": "TRAILING_STOP_MARKET",
                "activatePrice": "9020",            // activation price, only return with TRAILING_STOP_MARKET order
                "priceRate": "0.3",                 // callback rate, only return with TRAILING_STOP_MARKET order
                "updateTime": 1571110484038,
                "workingType": "CONTRACT_PRICE",
                "priceProtect": false            // if conditional order trigger is protected
            },
            {
                "code": -2011,
                "msg": "Unknown order sent."
            }
        ]
        """
        params = {
            "symbol": symbol,
            "timestamp": utils.get_current_timestamp()
        }
        response = self._signed_request(self._path_config.get("cancel_orders").get("method"),
                                        self._path_config.get("cancel_orders").get("path"),
                                        params)
        return response

    def get_data(self, symbol: str, interval: int, **kwargs):
        """
        API to get OHLCV data
        https://binance-docs.github.io/apidocs/spot/en/#uiklines

        :param symbol: future symbol
        :param interval: time interval
        :keyword start_date: start time of the data
        :keyword end_date: end time of the data
        :keyword limit: number of data limit
        :keyword is_dataframe: convert the data to pandas dataframe
        :return: list of list/ pandas dataframe
        """
        params = {
            "symbol": symbol,
            "interval": interval,
            "limit": kwargs.get("limit") if kwargs.get("limit", None) else self._LIMIT,
        }
        if kwargs.get("start_date"):
            params["startTime"] = int(kwargs["start_date"] * 1000)
        if kwargs.get("end_date"):
            params["endTime"] = int(kwargs["end_date"] * 1000)

        headers = {"X-MBX-APIKEY": self.key}
        response = self.send_request(self._path_config.get("get_data").get("method"),
                                     self._path_config.get("get_data").get("path"),
                                     headers, params)
        if kwargs.get("is_dataframe"):
            try:
                columns = ['Open Time', 'Open', 'High', 'Low', 'Close', 'Volume', 'Close time',
                           'Quote asset volume', 'Number of trades', 'Taker buy base asset volume',
                           'Taker buy quote asset volume', 'ignore']
                return pd.DataFrame(response, columns=columns)
            except Exception as e:
                logging.error("failed to create dataframe: ", e)
        return response

    def get_orderbook(self, symbol: str, **kwargs):
        """
        Get orderbook
        https://binance-docs.github.io/apidocs/spot/en/#recent-trades-list

        :param symbol: future symbol
        :keyword limit: result limit
        :keyword is_dataframe: convert the data to pandas dataframe
        :return: {
              "lastUpdateId": 1027024,
              "E": 1589436922972,   // Message output time
              "T": 1589436922959,   // Transaction time
              "bids": [
                [
                  "4.00000000",     // PRICE
                  "431.00000000"    // QTY
                ]
              ],
              "asks": [
                [
                  "4.00000200",
                  "12.00000000"
                ]
              ]
            }
        """
        params = {
            "symbol": symbol,
            "limit": kwargs.get("limit") if kwargs.get("limit") else self._LIMIT
        }
        header = {"X-MBX-APIKEY": self.key}
        response = super().send_request(self._path_config.get("get_orderbook").get("method"),
                                        self._path_config.get("get_orderbook").get("path"),
                                        header, params=params)
        if kwargs.get("is_dataframe"):
            try:
                columns = ['price', 'QTY']
                bid = pd.DataFrame(response["bids"], columns=columns)
                bid["type"] = ["bid" for i in range(0, bid.shape[0])]
                ask = pd.DataFrame(response["asks"], columns=columns)
                ask["type"] = ["ask" for i in range(0, ask.shape[0])]
                df = pd.concat([bid, ask], ignore_index=True)
                return df
            except Exception as e:
                logging.error("failed to create dataframe: ", e)
        return response

    def get_balance(self):
        """
        API to get future account balance
        https://binance-docs.github.io/apidocs/spot/en/#account-trade-list-user_data

        :return: [
                    {
                        "accountAlias": "SgsR",    // unique account code
                        "asset": "USDT",    // asset name
                        "balance": "122607.35137903", // wallet balance
                        "crossWalletBalance": "23.72469206", // crossed wallet balance
                        "crossUnPnl": "0.00000000"  // unrealized profit of crossed positions
                        "availableBalance": "23.72469206",       // available balance
                        "maxWithdrawAmount": "23.72469206",     // maximum amount for transfer out
                        "marginAvailable": true,    // whether the asset can be used as margin in Multi-Assets mode
                        "updateTime": 1617939110373
                    }
                ]
        """
        params = {
            "timestamp": utils.get_current_timestamp()
        }
        response = self._signed_request(self._path_config.get("get_balance").get("method"),
                                        self._path_config.get("get_balance").get("path"),
                                        params)
        return response

    def market_order(self, symbol: str, side: str, quantity: float):
        """
        API to place market order
        https://binance-docs.github.io/apidocs/spot/en/#new-order-trade

        :param symbol: future symbol
        :param side: buy/sell
        :param quantity:
        :return: {
            "clientOrderId": "testOrder",
            "cumQty": "0",
            "cumQuote": "0",
            "executedQty": "0",
            "orderId": 22542179,
            "avgPrice": "0.00000",
            "origQty": "10",
            "price": "0",
            "reduceOnly": false,
            "side": "BUY",
            "positionSide": "SHORT",
            "status": "NEW",
            "stopPrice": "9300",        // please ignore when order type is TRAILING_STOP_MARKET
            "closePosition": false,   // if Close-All
            "symbol": "BTCUSDT",
            "timeInForce": "GTC",
            "type": "TRAILING_STOP_MARKET",
            "origType": "TRAILING_STOP_MARKET",
            "activatePrice": "9020",    // activation price, only return with TRAILING_STOP_MARKET order
            "priceRate": "0.3",         // callback rate, only return with TRAILING_STOP_MARKET order
            "updateTime": 1566818724722,
            "workingType": "CONTRACT_PRICE",
            "priceProtect": false            // if conditional order trigger is protected
        }
        """
        params = {
            "symbol": symbol,
            "side": side.upper(),
            "type": "MARKET",
            "quantity": quantity,
            "timestamp": utils.get_current_timestamp()
        }
        response = self._signed_request(self._path_config.get("market_order").get("method"),
                                        self._path_config.get("market_order").get("path"),
                                        params)
        return response

    def limit_order(self, symbol: str, side: str, quantity: float, price: float, time_in_force="GTC"):
        """
        API to place limit order
        https://binance-docs.github.io/apidocs/spot/en/#new-order-trade

        :param symbol: future symbol
        :param side: buy/sell
        :param quantity: trade quantity
        :param price: trading price
        :param time_in_force: for postOnly request
        :return: {
            "clientOrderId": "testOrder",
            "cumQty": "0",
            "cumQuote": "0",
            "executedQty": "0",
            "orderId": 22542179,
            "avgPrice": "0.00000",
            "origQty": "10",
            "price": "0",
            "reduceOnly": false,
            "side": "BUY",
            "positionSide": "SHORT",
            "status": "NEW",
            "stopPrice": "9300",        // please ignore when order type is TRAILING_STOP_MARKET
            "closePosition": false,   // if Close-All
            "symbol": "BTCUSDT",
            "timeInForce": "GTC",
            "type": "TRAILING_STOP_MARKET",
            "origType": "TRAILING_STOP_MARKET",
            "activatePrice": "9020",    // activation price, only return with TRAILING_STOP_MARKET order
            "priceRate": "0.3",         // callback rate, only return with TRAILING_STOP_MARKET order
            "updateTime": 1566818724722,
            "workingType": "CONTRACT_PRICE",
            "priceProtect": false            // if conditional order trigger is protected
        }
        """
        params = {
            "symbol": symbol,
            "side": side.upper(),
            "type": "LIMIT",
            "quantity": quantity,
            "price": price,
            "timeInForce": time_in_force,
            "timestamp": utils.get_current_timestamp()
        }
        response = self._signed_request(self._path_config.get("limit_order").get("method"),
                                        self._path_config.get("limit_order").get("path"),
                                        params)
        return response

    def get_closed_orders(self, symbol: str):
        """
        API to get all the closed orders
        https://binance-docs.github.io/apidocs/spot/en/#new-oco-trade

        :param symbol: future symbol
        :return: [
              {
                "avgPrice": "0.00000",
                "clientOrderId": "abc",
                "cumQuote": "0",
                "executedQty": "0",
                "orderId": 1917641,
                "origQty": "0.40",
                "origType": "TRAILING_STOP_MARKET",
                "price": "0",
                "reduceOnly": false,
                "side": "BUY",
                "positionSide": "SHORT",
                "status": "NEW",
                "stopPrice": "9300",                // please ignore when order type is TRAILING_STOP_MARKET
                "closePosition": false,   // if Close-All
                "symbol": "BTCUSDT",
                "time": 1579276756075,              // order time
                "timeInForce": "GTC",
                "type": "TRAILING_STOP_MARKET",
                "activatePrice": "9020",            // activation price, only return with TRAILING_STOP_MARKET order
                "priceRate": "0.3",                 // callback rate, only return with TRAILING_STOP_MARKET order
                "updateTime": 1579276756075,        // update time
                "workingType": "CONTRACT_PRICE",
                "priceProtect": false            // if conditional order trigger is protected
              }
            ]
        """
        params = {
            "symbol": symbol,
            "timestamp": utils.get_current_timestamp()
        }
        response = self._signed_request(self._path_config.get("get_closed_orders").get("method"),
                                        self._path_config.get("get_closed_orders").get("path"),
                                        params)
        if type(response) == list:
            return list(filter(lambda order: order["status"] == "FILLED", response))
        return response

    def get_open_orders(self, symbol: str):
        """
        API to get all active orders
        https://binance-docs.github.io/apidocs/spot/en/#all-orders-user_data

        :param symbol: symbol
        :return: [
              {
                "avgPrice": "0.00000",
                "clientOrderId": "abc",
                "cumQuote": "0",
                "executedQty": "0",
                "orderId": 1917641,
                "origQty": "0.40",
                "origType": "TRAILING_STOP_MARKET",
                "price": "0",
                "reduceOnly": false,
                "side": "BUY",
                "positionSide": "SHORT",
                "status": "NEW",
                "stopPrice": "9300",                // please ignore when order type is TRAILING_STOP_MARKET
                "closePosition": false,   // if Close-All
                "symbol": "BTCUSDT",
                "time": 1579276756075,              // order time
                "timeInForce": "GTC",
                "type": "TRAILING_STOP_MARKET",
                "activatePrice": "9020",            // activation price, only return with TRAILING_STOP_MARKET order
                "priceRate": "0.3",                 // callback rate, only return with TRAILING_STOP_MARKET order
                "updateTime": 1579276756075,        // update time
                "workingType": "CONTRACT_PRICE",
                "priceProtect": false            // if conditional order trigger is protected
              }
            ]
        """
        params = {
            "symbol": symbol,
            "timestamp": utils.get_current_timestamp()
        }
        response = self._signed_request(self._path_config.get("get_open_orders").get("method"),
                                        self._path_config.get("get_open_orders").get("path"),
                                        params)
        return response

    def _signed_request(self, method, url, params):
        headers = {"X-MBX-APIKEY": self.key}
        signature = self._get_request_credentials(params)
        params["signature"] = signature
        response = self.send_request(method, url, headers, urlencode(params))
        return response

    def _get_sign(self, data):
        m = hmac.new(self.secret.encode("utf-8"), str(data).encode("utf-8"), hashlib.sha256)
        return m.hexdigest()

    def _get_request_credentials(self, params):
        sign = self._get_sign(urlencode(params))
        return sign



import hmac
import hashlib
import json
import logging
import uuid

import pandas as pd
from urllib.parse import urlencode

from onecall.base import utils
from onecall.base.exchange import Exchange
from onecall.base import urls


class Phemex(Exchange):
    """
    Phemex API class
    https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md
    """
    # Interval constants
    MINUTE_1 = 60
    MINUTE_5 = 300
    MINUTE_15 = 900
    MINUTE_30 = 1800
    HOUR_1 = 3600
    HOUR_4 = 14400
    DAY_1 = 86400
    WEEK_1 = 604800
    MONTH_1 = 2592000
    SEASON_1 = 7776000
    YEAR_1 = 31104000

    # Limit constants
    LIMIT_5 = 5
    LIMIT_10 = 10
    LIMIT_50 = 50
    LIMIT_100 = 100
    LIMIT_500 = 500
    LIMIT_1000 = 1000

    _LIMIT = 500

    # Constants for Order side
    BUY_SIDE = 'Buy'
    SELL_SIDE = 'Sell'

    def __init__(self, key=None, secret=None, debug=False, **kwargs):
        """
        Phemex API class
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md

        :param key: key
        :param secret: secret key
        :param debug: connect to testnet
        """
        self._path_config = {
            "get_positions": {"method": "GET", "path": "/accounts/accountPositions", "rate_limit": 50},
            "cancel_orders": {"method": "DELETE", "path": "/orders/all", "rate_limit": 50},
            "get_data": {"method": "GET", "path": "/exchange/public/md/v2/kline", "rate_limit": 50},
            "get_orderbook": {"method": "GET", "path": "/md/orderbook", "rate_limit": 50},
            "get_balance": {"method": "GET", "path": "/accounts/accountPositions", "rate_limit": 50},
            "market_order": {"method": "POST", "path": "/orders", "rate_limit": 50},
            "limit_order": {"method": "POST", "path": "/orders", "rate_limit": 50},
            "get_closed_orders": {"method": "GET", "path": "/exchange/order/list", "rate_limit": 50},
            "get_open_orders": {"method": "GET", "path": "/orders/activeList", "rate_limit": 50}
        }

        if not debug:
            kwargs["base_url"] = urls.PHEMEX_FUT_BASE_URL
        else:
            kwargs["base_url"] = urls.PHEMEX_FUT_TEST_BASE_URL
        super().__init__(key, secret, **kwargs)
        return

    def get_positions(self, currency="USD"):
        """
        API to get current positions
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#querytradeaccount

        :param currency: symbol
        :return: {
                "code": 0,
                    "msg": "",
                    "data": {
                        "positions": [
                            {
                                "accountID": 0,
                                "symbol": "BTCUSD",
                                "currency": "BTC",
                                "side": "None",
                                "positionStatus": "Normal",
                                "crossMargin": false,
                                "leverageEr": 0,
                                "leverage": 0,
                                "initMarginReqEr": 0,
                                "initMarginReq": 0.01,
                                "maintMarginReqEr": 500000,
                                "maintMarginReq": 0.005,
                                "riskLimitEv": 10000000000,
                                "riskLimit": 100,
                                "size": 0,
                                "value": 0,
                                "valueEv": 0,
                                ...
                            }
                        ]
                    }
            }
        """
        params = {
            "currency": currency
        }
        response = self._signed_request(self._path_config.get("get_positions").get("method"),
                                        self._path_config.get("get_positions").get("path"),
                                        params)
        if response.get("data", None):
            return response.get("data", {}).get("positions")
        else:
            self._logger.error(response)
            return response

    def cancel_orders(self, symbol: str):
        """
        API to cancel all orders
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#cancelall

        :param symbol: future symbol
        :return: data part of response is subject to change
        """
        params = {
            "symbol": symbol,
            "untriggered": True,
        }
        response = self._signed_request(self._path_config.get("cancel_orders").get("method"),
                                        self._path_config.get("cancel_orders").get("path"),
                                        params)
        return response

    def get_data(self, symbol: str, interval: int, **kwargs):
        """
        API to get OHLCV data
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#querykline

        :param symbol: future symbol
        :param interval: interval
        :keyword limit: data limit
        :keyword is_dataframe: convert the data to pandas dataframe
        :return: {
            "code": 0,
            "msg": "OK",
            "data": {
                    "total": -1,
                    "rows": [
                        [<timestamp>, <interval>, <last_close>, <open>, <high>, <low>, <close>, <volume>, <turnover>],
                        ]
                }
            }
        """
        params = {
            "symbol": symbol,
            "resolution": interval,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("get_data").get("method"),
                                        self._path_config.get("get_data").get("path"),
                                        params)
        if kwargs.get("is_dataframe", None):
            try:
                columns = ["timestamp", "interval", "last_close", "open", "high", "low", "close", "volume", "turnover"]
                return pd.DataFrame(response["data"]["rows"], columns=columns)
            except Exception as e:
                self._logger.error(e)
        return response

    def get_orderbook(self, symbol: str, is_dataframe=False):
        """
        API to get orderbook
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#queryorderbook

        :param symbol: future_symbol
        :param is_dataframe: whether to return row json/dataframe
        :return: {
              "error": null,
              "id": 0,
              "result": {
                "book": {
                  "asks": [[<priceEp>, size>],],
                  "bids": [[<priceEp>, <size>],],
                },
                "depth": 30,
                "sequence": <sequence>,
                "timestamp": <timestamp>,
                "symbol": "<symbol>",
                "type": "snapshot"
              }
            }
        """
        params = {
            "symbol": symbol
        }
        response = self._signed_request(self._path_config.get("get_orderbook").get("method"),
                                        self._path_config.get("get_orderbook").get("path"),
                                        params)
        if is_dataframe and response.get("result"):
            try:
                columns = ['price', 'QTY']
                bid = pd.DataFrame(response["result"]["book"]["bids"], columns=columns)
                bid["type"] = ["bid" for i in range(0, bid.shape[0])]
                ask = pd.DataFrame(response["result"]["book"]["asks"], columns=columns)
                ask["type"] = ["ask" for i in range(0, ask.shape[0])]
                df = pd.concat([bid, ask], ignore_index=True)
                return df
            except Exception as e:
                logging.error(e)
        return response

    def get_balance(self, currency="USD"):
        """
        API to get account balance
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#querytradeaccount

        :param currency: currency. USD,BTC
        :return: {
            "code": 0,
                "msg": "",
                "data": {
                    "account": {
                        "accountId": 0,
                        "currency": "BTC",
                        "accountBalanceEv": 0,
                        "totalUsedBalanceEv": 0
                    },
                }
            }
        """
        params = {
            "currency": currency
        }
        response = self._signed_request(self._path_config.get("get_balance").get("method"),
                                        self._path_config.get("get_balance").get("path"),
                                        params)
        if response.get("data"):
            balance = response.get("data", {}).get("account")
            if currency == 'USD':
                balance["balance"] = balance.get("accountBalanceEv", 0) / 10000
            elif currency.upper() == 'BTC':
                balance["balance"] = balance.get("accountBalanceEv", 0) / 100000000
            return balance
        return response

    def market_order(self, symbol: str, side: str, order_qty: float, **kwargs):
        """
        API to place market order
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#placeorder

        :param symbol: coin symbol
        :param side: BUY/SELL
        :param order_qty: order quantity
        :keyword timeInForce: Time in force. default to GoodTillCancel
        :keyword reduceOnly: whether reduce position side only
        :return: {
                "code": 0,
                    "msg": "",
                    "data": {
                        "bizError": 0,
                        "orderID": "ab90a08c-b728-4b6b-97c4-36fa497335bf",
                        "clOrdID": "137e1928-5d25-fecd-dbd1-705ded659a4f",
                        "symbol": "BTCUSD",
                        "side": "Sell",
                        "actionTimeNs": 1580547265848034600,
                        "transactTimeNs": 0,
                        "orderType": null,
                        "priceEp": 98970000,
                        "price": 9897,
                        "orderQty": 1,
                        "displayQty": 1,
                        "timeInForce": null,
                        "reduceOnly": false,
                        "stopPxEp": 0,
                        "closedPnlEv": 0,
                        "closedPnl": 0,
                        "closedSize": 0,
                        "cumQty": 0,
                        "cumValueEv": 0,
                        "cumValue": 0,
                        "leavesQty": 1,
                        "leavesValueEv": 10104,
                        "leavesValue": 0.00010104,
                        "stopPx": 0,
                        "stopDirection": "UNSPECIFIED",
                        "ordStatus": "Created"
                    }
            }
        """
        payload = {
            "clOrdID": str(uuid.uuid1()),
            "symbol": symbol,
            "side": side,
            "ordType": "Market",
            "orderQty": order_qty,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("market_order").get("method"),
                                        self._path_config.get("market_order").get("path"),
                                        data=payload)
        return response

    def limit_order(self, symbol: str, side: str, order_qty: int, price: float, **kwargs):
        """
        API to place limit order
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#placeorder

        :param symbol: currency symbol
        :param side: BUY/SELL
        :param order_qty: order quantity
        :param price: order price
        :keyword timeInForce: Time in force. default to GoodTillCancel
        :keyword reduceOnly: whether reduce position side only
        :return: {
                "code": 0,
                    "msg": "",
                    "data": {
                        "bizError": 0,
                        "orderID": "ab90a08c-b728-4b6b-97c4-36fa497335bf",
                        "clOrdID": "137e1928-5d25-fecd-dbd1-705ded659a4f",
                        "symbol": "BTCUSD",
                        "side": "Sell",
                        "actionTimeNs": 1580547265848034600,
                        "transactTimeNs": 0,
                        "orderType": null,
                        "priceEp": 98970000,
                        "price": 9897,
                        "orderQty": 1,
                        "displayQty": 1,
                        "timeInForce": null,
                        "reduceOnly": false,
                        "stopPxEp": 0,
                        "closedPnlEv": 0,
                        "closedPnl": 0,
                        "closedSize": 0,
                        "cumQty": 0,
                        "cumValueEv": 0,
                        "cumValue": 0,
                        "leavesQty": 1,
                        "leavesValueEv": 10104,
                        "leavesValue": 0.00010104,
                        "stopPx": 0,
                        "stopDirection": "UNSPECIFIED",
                        "ordStatus": "Created"
                    }
            }
        """
        payload = {
            "clOrdID": str(uuid.uuid1()),
            "symbol": symbol,
            "side": side,
            "ordType": "Limit",
            "orderQty": order_qty,
            "priceEp": price,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("limit_order").get("method"),
                                        self._path_config.get("limit_order").get("path"),
                                        data=payload)
        return response

    def get_closed_orders(self, symbol):
        """
        API to get closed order
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#queryorder

        :param symbol: currency symbol
        :return: {
                "code": 0,
                    "msg": "OK",
                    "data": {
                        "total": 39,
                        "rows": [
                        {
                            "orderID": "7d5a39d6-ff14-4428-b9e1-1fcf1800d6ac",
                            "clOrdID": "e422be37-074c-403d-aac8-ad94827f60c1",
                            "symbol": "BTCUSD",
                            "side": "Sell",
                            "orderType": "Limit",
                            "actionTimeNs": 1577523473419470300,
                            "priceEp": 75720000,
                            "price": null,
                            "orderQty": 12,
                            "displayQty": 0,
                            "timeInForce": "GoodTillCancel",
                            "reduceOnly": false,
                            "takeProfitEp": 0,
                            "takeProfit": null,
                            "stopLossEp": 0,
                            ...
                            "ordStatus": "Filled",
                        }
                    ]
                }
            }
        """
        params = {
            "symbol": symbol,
            "ordStatus": "Filled"
        }
        response = self._signed_request(self._path_config.get("get_closed_orders").get("method"),
                                        self._path_config.get("get_closed_orders").get("path"),
                                        params)
        return response

    def get_open_orders(self, symbol):
        """
        API to get all open orders by symbol
        https://github.com/phemex/phemex-api-docs/blob/master/Public-Contract-API-en.md#queryopenorder

        :param symbol: currency symbol
        :return: {
                "code": 0,
                    "msg": "",
                    "data": {
                        "rows": [
                        {
                            "bizError": 0,
                            "orderID": "9cb95282-7840-42d6-9768-ab8901385a67",
                            "clOrdID": "7eaa9987-928c-652e-cc6a-82fc35641706",
                            "symbol": "BTCUSD",
                            "side": "Buy",
                            "actionTimeNs": 1580533011677666800,
                            "transactTimeNs": 1580533011677666800,
                            "orderType": null,
                            "priceEp": 84000000,
                            "price": 8400,
                            "orderQty": 1,
                            "displayQty": 1,
                            "timeInForce": null,
                            "reduceOnly": false,
                            "stopPxEp": 0,
                            "closedPnlEv": 0,
                            "closedPnl": 0,
                            "closedSize": 0,
                            "cumQty": 0,
                            "cumValueEv": 0,
                            "cumValue": 0,
                            "leavesQty": 0,
                            "leavesValueEv": 0,
                            "leavesValue": 0,
                            "stopPx": 0,
                            "stopDirection": "Falling",
                            "ordStatus": "Untriggered"
                        }
                    ]
                }
            }
        """
        params = {
            "symbol": symbol,
        }
        response = self._signed_request(self._path_config.get("get_open_orders").get("method"),
                                        self._path_config.get("get_open_orders").get("path"),
                                        params)
        return response

    def _signed_request(self, method, path, params=None, data=None):
        expiry = utils.get_current_timestamp() + 1
        query_params = '&'.join(['{}={}'.format(k, v) for k, v in params.items()]) if params else None
        if data:
            data = json.dumps(data, separators=(',', ':')) if data else None
        header = self._get_request_credentials(path, expiry, params=query_params, payload=data)
        if data:
            data = data.encode()
        response = self.send_request(method, path, header, query_params, data)
        return response

    def _get_sign(self, data):
        m = hmac.new(self.secret.encode("utf-8"), data.encode("utf-8"), hashlib.sha256)
        return m.hexdigest()

    def _get_request_credentials(self, path, expiry, **kwargs):
        signature = ""
        if kwargs.get("params"):
            signature = self._get_sign(path + kwargs.get("params") + str(expiry))
        elif kwargs.get("payload"):
            signature = self._get_sign(path + str(expiry) + kwargs.get("payload"))
        header = {
            "x-phemex-access-token": self.key,
            "x-phemex-request-signature": signature,
            "x-phemex-request-expiry": str(expiry),
            "Content-Type": "application/json"
        }
        return header



import hmac
import hashlib
import logging
import pandas as pd
import base64
import uuid
import json
from urllib.parse import urljoin

from onecall.base import utils
from onecall.base.exchange import Exchange
from onecall.base import urls


class Kucoin(Exchange):
    """
    Kucoin API class
    """
    # interval
    INTERVAL_1 = 15
    INTERVAL_15 = 15
    INTERVAL_30 = 30
    INTERVAL_60 = 60
    INTERVAL_120 = 120
    INTERVAL_240 = 240
    INTERVAL_480 = 480
    INTERVAL_720 = 720
    INTERVAL_1440 = 1440
    INTERVAL_10080 = 10080

    # Constants for Order side
    BUY_SIDE = "buy"
    SELL_SIDE = "sell"

    def __init__(self, key=None, secret=None, passphrase=None, debug=False, **kwargs):
        """
        Kucoin API class
        https://docs.kucoin.com/futures/#general

        :param key: api key
        :param secret: secret key
        :param passphrase: passphrase
        :param debug: flag to switch to test env
        """
        self._path_config = {
            "get_positions": {"method": 'GET', "path": "/api/v1/position", "rate_limit": 50},
            "cancel_orders": {"method": "DELETE", "path": "/api/v1/orders", "rate_limit": 50},
            "get_data": {"method": "GET", "path": "/api/v1/kline/query", "rate_limit": 50},
            "get_orderbook": {"method": "GET", "path": "/api/v1/level2/snapshot", "rate_limit": 50},
            "get_balance": {"method": "GET", "path": "/api/v1/account-overview", "rate_limit": 50},
            "market_order": {"method": "POST", "path": "/api/v1/orders", "rate_limit": 50},
            "limit_order": {"method": "POST", "path": "/api/v1/orders", "rate_limit": 50},
            "get_closed_orders": {"method": "GET", "path": "/api/v1/orders", "rate_limit": 50},
            "get_open_orders": {"method": "GET", "path": "/api/v1/orders", "rate_limit": 50}
        }
        if not debug:
            kwargs["base_url"] = urls.KUCOIN_FUT_BASE_URL
        else:
            kwargs["base_url"] = urls.KUCOIN_FUT_TEST_BASE_URL
        super().__init__(key, secret, passphrase, **kwargs)
        return

    def get_positions(self, symbol: str):
        """
        API to get current positions in future
        https://docs.kucoin.com/futures/?lang=en_US#get-position-details

        :param symbol: symbol
        :return: {
            "code": "200000",
            "data": {
            "id": "5e81a7827911f40008e80715",//Position ID
            "symbol": "XBTUSDTM",//Symbol
            "autoDeposit": False,//Auto deposit margin or not
            "maintMarginReq": 0.005,//Maintenance margin requirement
            "riskLimit": 2000000,//Risk limit
            "realLeverage": 5.0,//Leverage o the order
            "crossMode": False,//Cross mode or not
            "delevPercentage": 0.35,//ADL ranking percentile
            "openingTimestamp": 1623832410892,//Open time
            "currentTimestamp": 1623832488929,//Current timestamp
            "currentQty": 1,//Current postion quantity
            ...
            }
        }
        """
        params = {
            "symbol": symbol
        }
        response = self.__signed_request(self._path_config.get("get_positions").get("method"),
                                         self._path_config.get("get_positions").get("path"),
                                         params)
        return response

    def cancel_orders(self, symbol: str):
        """
        API to cancel all the active orders
        https://docs.kucoin.com/futures/?lang=en_US#limit-order-mass-cancelation

        :param symbol: symbol
        :return:  {
            "code": "200000",
            "data": {
              "cancelledOrderIds": [
                "5c52e11203aa677f33e493fb",
                "5c52e12103aa677f33e493fe",
              ]
            }
          }
        """
        params = {
            "symbol": symbol
        }
        response = self.__signed_request(self._path_config.get("cancel_orders").get("method"),
                                         self._path_config.get("cancel_orders").get("path"),
                                         params)
        return response

    def get_data(self, symbol: str, interval: int, **kwargs):
        """
        API to get OHLCV data
        https://docs.kucoin.com/futures/?lang=en_US#get-k-line-data-of-contract

        :param symbol: symbol
        :param interval: chart interval
        :keyword from: start time
        :keyword ti: end time
        :keyword is_dataframe: convert data to dataframe or not
        :return: list of list/ pandas dataframe
        """
        params = {
            "symbol": symbol,
            "granularity": interval,
            **kwargs
        }
        response = self.__signed_request(self._path_config.get("get_data").get("method"),
                                         self._path_config.get("get_data").get("path"),
                                         params)
        if kwargs.get("is_dataframe", None) and response.get("data"):
            try:
                columns = ['Time', 'Entry price', 'Highest price', 'Lowest price', 'Close price', 'Volume']
                return pd.DataFrame(response["data"], columns=columns)
            except Exception as e:
                logging.error("failed to create dataframe : ", e)
        return response

    def get_orderbook(self, symbol: str, **kwargs):
        """
        Get orderbook
        https://docs.kucoin.com/futures/?lang=en_US#order-book

        :param symbol: contract symbol
        :keyword is_dataframe: convert data to dataframe or not
        :return: list of list/ pandas dataframe
        """
        params = {
            "symbol": symbol,
            **kwargs
        }
        response = self.__signed_request(self._path_config.get("get_orderbook").get("method"),
                                         self._path_config.get("get_orderbook").get("path"),
                                         params)
        if kwargs.get("is_dataframe", None) and response.get("data"):
            try:
                columns = ['price', 'QTY']
                bid = pd.DataFrame(response["data"]["bids"], columns=columns)
                bid["type"] = ["bid" for i in range(0, bid.shape[0])]
                ask = pd.DataFrame(response["data"]["asks"], columns=columns)
                ask["type"] = ["ask" for i in range(0, ask.shape[0])]
                df = pd.concat([bid, ask], ignore_index=True)
                return df
            except Exception as e:
                logging.error("failed to create dataframe: ", e)
        return response

    def get_balance(self):
        """
        API to get future account balance
        https://docs.kucoin.com/futures/?lang=en_US#get-account-overview

        :return: list
        """
        response = self.__signed_request(self._path_config.get("get_balance").get("method"),
                                         self._path_config.get("get_balance").get("path"))
        return response

    def market_order(self, symbol: str, side: str, quantity: float, leverage="5", **kwargs):
        """
        API to place market order
        https://docs.kucoin.com/futures/?lang=en_US#place-an-order

        :param symbol: symbol
        :param side: buy/sell
        :param quantity: order quantity
        :param leverage: Leverage of the order
        :return: {
            "code": "200000",
            "data": {
              "orderId": "5bd6e9286d99522a52e458de"
              }
          }
        """
        payload = {
            "clientOid": str(uuid.uuid1()),
            "symbol": symbol,
            "side": side,
            "type": "market",
            "size": quantity,
            "leverage": leverage,
            **kwargs
        }
        response = self.__signed_request(self._path_config.get("market_order").get("method"),
                                         self._path_config.get("market_order").get("path"),
                                         data=payload)
        return response

    def limit_order(self, symbol: str, side: str, price: str, quantity: int, **kwargs):
        """
        API to place limit order
        https://docs.kucoin.com/futures/?lang=en_US#place-an-order

        :param symbol: symbol
        :param side: buy/sell
        :param quantity: order quantity
        :param price: order price
        :return: {
            "code": "200000",
            "data": {
              "orderId": "5bd6e9286d99522a52e458de"
              }
          }
        """
        payload = {
            "clientOid": str(uuid.uuid1()),
            "symbol": symbol,
            "side": side,
            "price": price,
            "type": "limit",
            "size": quantity,
            **kwargs
        }
        response = self.__signed_request(self._path_config.get("limit_order").get("method"),
                                         self._path_config.get("limit_order").get("path"),
                                         data=payload)
        return response

    def get_closed_orders(self, **kwargs):
        """
        API to get closed orders
        https://docs.kucoin.com/futures/?lang=en_US#get-order-list

        :return: {
                "code": "200000",
                "data": {
                  "currentPage":1,
                  "pageSize":1,
                  "totalNum":251915,
                  "totalPage":251915,
                  "items":[
                      {
                        "symbol": "XBTUSDM",  //Symbol of the contract
                        "tradeId": "5ce24c1f0c19fc3c58edc47c",  //Trade ID
                        "orderId": "5ce24c16b210233c36ee321d",  // Order ID
                        "side": "sell",  //Transaction side
                        "liquidity": "taker",  //Liquidity- taker or maker
                        "forceTaker": true, //Whether to force processing as a taker
                        "price": "8302",  //Filled price
                        "size": 10,  //Filled amount
                        "value": "0.001204529",  //Order value
                        ...
                      }
                  ]
                }
            }
        """
        params = {
            "status": "done",
            **kwargs
        }
        response = self.__signed_request(self._path_config.get("get_closed_orders").get("method"),
                                         self._path_config.get("get_closed_orders").get("path"),
                                         params)
        return response

    def get_open_orders(self, **kwargs):
        """
        API to get open orders
        https://docs.kucoin.com/futures/?lang=en_US#get-order-list

        :return:
        """
        params = {
            "status": "active",
            **kwargs
        }
        response = self.__signed_request(self._path_config.get("get_open_orders").get("method"),
                                         self._path_config.get("get_open_orders").get("path"),
                                         params)
        return response

    def __signed_request(self, method, url, params=None, data=None):
        param_data = ""
        if params:
            strl = []
            for key in sorted(params):
                strl.append("{}={}".format(key, params[key]))
            param_data += '&'.join(strl)
            url += '?' + param_data
        if data:
            data = json.dumps(data)
        header = self._get_request_credentials(method, url, data=data)
        response = self.send_request(method, url, header, data=data)
        return response

    def _get_sign(self, str_to_sign):
        signature = base64.b64encode(hmac.new(self.secret.encode('utf-8'), str_to_sign.encode('utf-8'),
                                              hashlib.sha256).digest())
        passphrase = base64.b64encode(hmac.new(self.secret.encode('utf-8'), self.pass_phrase.encode('utf-8'),
                                               hashlib.sha256).digest())
        return signature, passphrase

    def _get_request_credentials(self, method, path, **kwargs):
        timestamp = utils.get_current_timestamp()
        if kwargs.get("data"):
            sign_string = str(timestamp)+method+path+kwargs["data"]
        else:
            sign_string = str(timestamp)+method+path
        sign, pass_phrase = self._get_sign(sign_string)
        headers = {
            "KC-API-SIGN": sign,
            "KC-API-TIMESTAMP": str(timestamp),
            "KC-API-KEY": self.key,
            "KC-API-PASSPHRASE": pass_phrase,
            "KC-API-KEY-VERSION": "2",
            "Content-Type": "application/json"
        }
        return headers



from onecall.base.exchange import Exchange

from onecall.binance import Binance as binance
from onecall.binance_spot import BinanceSpot as binancespot
from onecall.phemex import Phemex as phemex
from onecall.kucoin import Kucoin as kucoin
from onecall.bybit import Bybit as bybit
from onecall.ftx_us import FTXUS as ftxus
from onecall.ftx import FTX as ftx

exchanges = [
    "Biance",
    'Phemex',
    'Kucoin',
    'Bybit',
    'FTX',
    'Binance_spot',
]

__all__ = [
    binance,
    binancespot,
    phemex,
    kucoin,
    bybit,
    ftxus,
    ftx,
    exchanges,
]



from onecall.ftx import FTX
from onecall.base import urls


class FTXUS(FTX):
    """
    API class for FTX.US
    """
    def __init__(self, key=None, secret=None, **kwargs):
        """
        FTX.US API class
        https://docs.ftx.com/#rest-api

        :param key: API key
        :param secret: Secret key
        """
        kwargs["base_url"] = urls.FTXUS_FUT_BASE_URL

        super().__init__(key, secret, **kwargs)
        return

    def get_positions(self, symbol):
        return {"error": "spot market don't have position concept. check balance instead"}



from json import JSONDecodeError
import requests
import logging
import json

from .exceptions import ClientException, ServerException


class Exchange:
    """
    API Basic class
    """
    def __init__(
            self,
            key=None,
            secret=None,
            pass_phrase=None,
            base_url=None,
            show_limit_usage=False,
    ):
        """
        Initialise client

        :param key: API key
        :param secret: secret key
        :param pass_phrase: pass phrase for kucoin
        :param base_url: base url
        :param show_limit_usage: flag to show usage
        """
        self.key = key
        self.secret = secret
        self.pass_phrase = pass_phrase
        self.show_limit_usage = False
        self.session = requests.Session()

        if base_url:
            self.base_url = base_url

        if show_limit_usage is True:
            self.show_limit_usage = True

        self._logger = logging.getLogger(__name__)
        return

    def send_request(self, http_method, url_path, header=None, params=None, data=None):
        """
        function to create and send HTTP request

        :param http_method: http method
        :param url_path: api uri path
        :param header: request header
        :param params: request query parameters
        :param data: request json data
        :return: http response
        """
        try:
            url = self.base_url + url_path
            self._logger.debug("url: " + url)
            payload = {
                "url": url,
                "headers": header,
                "params": params,
                "data": data
            }
            response = self._dispatch_request(http_method)(**payload)
            self._logger.debug("raw response from server:" + response.text)
            self._handle_exception(response)
            return response.json()
        except Exception as e:
            return self._send_error_response(e)

    def _dispatch_request(self, http_method):
        """
        function to get http dispatch method

        :param http_method: http method
        :return: http method
        """
        return {
            "GET": self.session.get,
            "DELETE": self.session.delete,
            "PUT": self.session.put,
            "POST": self.session.post,
        }.get(http_method, "GET")

    def _handle_exception(self, response):
        status_code = response.status_code
        if status_code < 400:
            return
        if 400 <= status_code < 500:
            try:
                err = json.loads(response.text)
            except JSONDecodeError:
                raise ClientException(response.text)
            raise ClientException(response.text)
        raise ServerException(response.text)

    def _send_error_response(self, e: Exception):
        return {"error": e.args}



from phemex import Phemex


def main():
    client = Phemex(key="2f5efa34-59e8-4eb8-95f1-821d72054fe9",
                         secret="WcrWD3PceZ4KaeiMx1_z0ecQZJ7-Yq7CUbeNGaX7SrJjOGIyYzk0Yi03YzUyLTRlYmEtYWJiZi00Y2RjZjMxMzk0MDQ",
                         debug=True)

    print(client.get_balance('BTC'))


if __name__ == "__main__":
    main()



import hmac
import hashlib
import json
import logging
import pandas as pd
from urllib.parse import urlencode

from onecall.base import utils
from onecall.base.exchange import Exchange
from onecall.base import urls


class Bybit(Exchange):
    """
    Bybit API Class
    """
    INTERVAL_1m = '1'
    INTERVAL_3m = '3'
    INTERVAL_5m = '5'
    INTERVAL_15m = '15'
    INTERVAL_30m = '30'
    INTERVAL_1H = '60'
    INTERVAL_2H = '120'
    INTERVAL_4H = '240'
    INTERVAL_6H = '360'
    INTERVAL_12H = '720'
    INTERVAL_1D = 'D'
    INTERVAL_1W = 'W'
    INTERVAL_1M = 'M'

    # Constants for Order side
    BUY_SIDE = 'Buy'
    SELL_SIDE = 'Sell'

    def __init__(self, key=None, secret=None, debug=False, **kwargs):
        """
        Bybit API class
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-introduction

        :param key: api key
        :param secret: secret key
        :param debug: flag to switch to test env
        """
        self._path_config = {
            "get_positions": {"method": "GET", "path": "/private/linear/position/list", "rate_limit": 50},
            "cancel_active_orders": {"method": "POST", "path": "/private/linear/order/cancel-all", "rate_limit": 50},
            "cancel_stop_orders": {"method": "POST", "path": "/private/linear/stop-order/cancel-all", "rate_limit": 50},
            "get_data": {"method": "GET", "path": "/public/linear/kline", "rate_limit": 50},
            "get_orderbook": {"method": "GET", "path": "/v2/public/orderBook/L2", "rate_limit": 50},
            "get_balance": {"method": "GET", "path": "/v2/private/wallet/balance", "rate_limit": 50},
            "market_order": {"method": "POST", "path": "/private/linear/order/create", "rate_limit": 50},
            "limit_order": {"method": "POST", "path": "/private/linear/order/create", "rate_limit": 50},
            "get_closed_orders": {"method": "GET", "path": "/private/linear/order/list", "rate_limit": 50},
            "get_open_orders": {"method": "GET", "path": "/private/linear/order/search", "rate_limit": 50}
        }
        self.recv_window = "5000"
        if not debug:
            kwargs["base_url"] = urls.BYBIT_FUT_BASE_URL
        else:
            kwargs["base_url"] = urls.BYBIT_FUT_TEST_BASE_URL
        super().__init__(key, secret, **kwargs)
        return

    def get_positions(self, symbol: str):
        """
        API to get current positions
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-position

        :param symbol: contract symbol
        :return: {
            "ret_code": 0,
            "ret_msg": "OK",
            "ext_code": "",
            "ext_info": "",
            "result": [
                {
                    "user_id": 533285,
                    "symbol": "XRPUSDT",
                    "side": "Buy",
                    "size": 0,
                    "position_value": 0,
                    "entry_price": 0,
                    "liq_price": 0,
                    "bust_price": 0,
                    "leverage": 5,
                    "auto_add_margin": 0,
                    "is_isolated": true,
                    "position_margin": 0,
                    }
                ]
            }
        """
        params = {
            "symbol": symbol,
        }
        response = self._signed_request(self._path_config.get("get_positions").get("method"),
                                        self._path_config.get("get_positions").get("path"),
                                        params)
        return response

    def cancel_orders(self, symbol: str):
        """
        API to cancel all orders
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-cancelallactive
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-cancelallcond

        :param symbol: symbol
        :return: {"active_order": {active_response}, "conditional_order": {conditional_response}}
        """
        payload = {
            "symbol": symbol,
        }
        active_response = self._signed_request(self._path_config.get("cancel_active_orders").get("method"),
                                               self._path_config.get("cancel_active_orders").get("path"),
                                               data=payload)
        conditional_response = self._signed_request(self._path_config.get("cancel_stop_orders").get("method"),
                                                    self._path_config.get("cancel_stop_orders").get("path"),
                                                    data=payload)
        return {"active_order": active_response, "conditional_order": conditional_response}

    def get_data(self, symbol: str, interval: str, start_time: int, **kwargs):
        """
        API to get OHLCV data
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-querykline

        :param symbol: symbol
        :param interval: interval
        :param start_time: start time
        :keyword limit: data limit
        :keyword is_dataframe: convert the data to pandas dataframe
        :return: list of list/ pandas dataframe
        """
        params = {
            "symbol": symbol,
            "interval": interval,
            "from": start_time,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("get_data").get("method"),
                                        self._path_config.get("get_data").get("path"),
                                        params)
        if kwargs.get("is_dataframe"):
            try:
                return pd.DataFrame(response["result"])
            except Exception as e:
                logging.error("failed to create dataframe: ", e)
        return response

    def get_orderbook(self, symbol, is_dataframe=False):
        """
        API to get orderbook
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-orderbook

        :param symbol: symbol
        :param is_dataframe: convert the data to pandas dataframe
        :return: {
                "ret_code": 0,
                "ret_msg": "OK",
                "ext_code": "",
                "ext_info": "",
                "result": [
                    {
                        "symbol": "BTCUSD",
                        "price": "9487",
                        "size": 336241,
                        "side": "Buy"
                    },
                    {
                        "symbol": "BTCUSD",
                        "price": "9487.5",
                        "size": 522147,
                        "side": "Sell"
                    }
                ],
                "time_now": "1567108756.834357"
            }
        """
        params = {
            "symbol": symbol
        }
        response = self._signed_request(self._path_config.get("get_orderbook").get("method"),
                                        self._path_config.get("get_orderbook").get("path"),
                                        params)
        if is_dataframe:
            try:
                return pd.DataFrame(response["result"])
            except Exception as e:
                logging.error("failed to create dataframe:", e)
        return response

    def get_balance(self):
        """
        API to get balance
        https://bybit-exchange.github.io/docs/futuresV2/inverse/#t-balance

        :return: {
                "ret_code": 0,
                "ret_msg": "OK",
                "ext_code": "",
                "ext_info": "",
                "result": {
                    "BTC": {
                        "equity": 1002,
                        "available_balance": 999.99987471,
                        "used_margin": 0.00012529,
                        "order_margin": 0.00012529,
                        "position_margin": 0,
                        "occ_closing_fee": 0,
                        "occ_funding_fee": 0,
                        "wallet_balance": 1000,
                        "realised_pnl": 0,
                        "unrealised_pnl": 2,
                        "cum_realised_pnl": 0,
                        "given_cash": 0,
                        "service_cash": 0
                    }
                },
                "time_now": "1578284274.816029",
                "rate_limit_status": 98,
                "rate_limit_reset_ms": 1580885703683,
                "rate_limit": 100
            }
        """
        response = self._signed_request(self._path_config.get("get_balance").get("method"),
                                        self._path_config.get("get_balance").get("path"))
        return response

    def market_order(self, symbol, side, quantity, time_in_force="GoodTillCancel", reduce_only=False,
                     close_on_trigger=False, **kwargs):
        """
        API to create market order
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-placeactive

        :param symbol: symbol
        :param side: buy/sell
        :param quantity: order quantity
        :param time_in_force: time in force
        :param reduce_only: your position can only reduce in size if this order is triggered
        :param close_on_trigger: For a closing order. It can only reduce your position, not increase it.
        :return: {
                "ret_code": 0,
                "ret_msg": "OK",
                "ext_code": "",
                "ext_info": "",
                "result": {
                    "user_id": 533285,
                    "order_id": "a1904030-f99c-4e35-9217-111591f08493",
                    "symbol": "BTCUSD",
                    "side": "Buy",
                    "order_type": "Limit",
                    "price": 20010,
                    "qty": 200,
                    ...
                },
                "time_now": "1655708182.071362",
                "rate_limit_status": 99,
                "rate_limit_reset_ms": 1655708182068,
                "rate_limit": 100
            }
        """
        payload = {
            "side": side,
            "symbol": symbol,
            "order_type": "Market",
            "qty": quantity,
            "time_in_force": time_in_force,
            "reduce_only": reduce_only,
            "close_on_trigger": close_on_trigger,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("market_order").get("method"),
                                        self._path_config.get("market_order").get("path"),
                                        data=payload)
        return response

    def limit_order(self, symbol, side, quantity, price, time_in_force="GoodTillCancel", reduce_only=False,
                    close_on_trigger=False, **kwargs):
        """
        API to create limit order
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-placeactive

        :param symbol: symbol
        :param side: buy/sell
        :param quantity: order quantity
        :param price: order price
        :param time_in_force: time in force
        :param reduce_only: your position can only reduce in size if this order is triggered
        :param close_on_trigger: For a closing order. It can only reduce your position, not increase it.
        :return: {
                    "user_id": 533285,
                    "order_id": "a1904030-f99c-4e35-9217-111591f08493",
                    "symbol": "BTCUSD",
                    "side": "Buy",
                    "order_type": "Limit",
                    "price": 20010,
                    "qty": 200,
                    ...
                },
                "time_now": "1655708182.071362",
                "rate_limit_status": 99,
                "rate_limit_reset_ms": 1655708182068,
                "rate_limit": 100
            }
        """
        payload = {
            "side": side,
            "symbol": symbol,
            "order_type": "Limit",
            "qty": quantity,
            "price": price,
            "time_in_force": time_in_force,
            "reduce_only": reduce_only,
            "close_on_trigger": close_on_trigger,
            **kwargs
        }
        response = self._signed_request(self._path_config.get("limit_order").get("method"),
                                        self._path_config.get("limit_order").get("path"),
                                        data=payload)
        return response

    def get_closed_orders(self, symbol):
        """
        API to get closed orders
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-getactive

        :param symbol: symbols
        :return: {
                "ret_code": 0,
                "ret_msg": "OK",
                "ext_code": "",
                "ext_info": "",
                "result": [
                    {
                        "order_id": "8e7d1cd2-d7b3-4c61-87a9-a85a6e59cef8",
                        "user_id": 533285,
                        "symbol": "BITUSDT",
                        "side": "Buy",
                        "order_type": "Limit",
                        "price": 0.3,
                        "qty": 100,
                        "time_in_force": "GoodTillCancel",
                        "order_status": "New",
                        "last_exec_price": 0,
                    }
                ],
                "time_now": "1655863947.736147",
                "rate_limit_status": 599,
                "rate_limit_reset_ms": 1655863947734,
                "rate_limit": 600
            }
        """
        param = {
            "symbol": symbol
        }
        response = self._signed_request(self._path_config.get("get_closed_orders").get("method"),
                                        self._path_config.get("get_closed_orders").get("path"),
                                        param)
        if response.get("result"):
            return list(filter(lambda order: order.get("order_status") == "Filled",
                               response.get("result", {}).get("data")))
        return response

    def get_open_orders(self, symbol):
        """
        API to get open orders
        https://bybit-exchange.github.io/docs/futuresV2/linear/#t-queryactive

        :param symbol: symbol
        :return: {
                "ret_code": 0,
                "ret_msg": "OK",
                "ext_code": "",
                "ext_info": "",
                "result": [
                    {
                        "order_id": "8e7d1cd2-d7b3-4c61-87a9-a85a6e59cef8",
                        "user_id": 533285,
                        "symbol": "BITUSDT",
                        "side": "Buy",
                        "order_type": "Limit",
                        "price": 0.3,
                        "qty": 100,
                        "time_in_force": "GoodTillCancel",
                        "order_status": "New",
                        "last_exec_price": 0,
                    }
                ],
                "time_now": "1655863947.736147",
                "rate_limit_status": 599,
                "rate_limit_reset_ms": 1655863947734,
                "rate_limit": 600
            }
        """
        param = {
            "symbol": symbol
        }
        response = self._signed_request(self._path_config.get("get_open_orders").get("method"),
                                        self._path_config.get("get_open_orders").get("path"),
                                        param)
        return response

    def _signed_request(self, method, url, params=None, data=None):
        timestamp = str(utils.get_current_timestamp())
        if data:
            data["timestamp"] = timestamp
            data["api_key"] = self.key
        elif params:
            params["timestamp"] = timestamp
            params["api_key"] = self.key
        else:
            params = {"timestamp": timestamp, "api_key": self.key}
        sign = self._get_request_credentials(params=params, data=data)
        headers = {"Content-Type": "application/json"}
        if data:
            data = json.dumps(dict(data, **sign))
        else:
            params = urlencode(params) + "&sign=" + sign.get("sign")
        response = self.send_request(method, url, headers, params, data)
        return response

    def _get_sign(self, str_to_sign):
        sign = hmac.new(self.secret.encode("utf-8"), str_to_sign.encode("utf-8"), hashlib.sha256)
        return sign.hexdigest()

    def _get_request_credentials(self, params=None, data=None):
        if data:
            sign = self._get_sign(self._get_sign_string(data))
        else:
            sign = self._get_sign(self._get_sign_string(params))
        return {"sign": sign}

    def _get_sign_string(self, params):
        sign_string = ""
        for key in sorted(params.keys()):
            v = params[key]
            if isinstance(params[key], bool):
                if params[key]:
                    v = 'true'
                else:
                    v = 'false'
            sign_string += key + '=' + str(v) + '&'
        return sign_string[:-1]



# Binance
BINANCE_SPOT_BASE_URL = "https://api.binance.com"
BINANCE_FUT_BASE_URL = "https://fapi.binance.com"
BINANCE_SPOT_TEST_BASE_URL = "https://testnet.binance.vision"
BINANCE_FUT_TEST_BASE_URL = "https://testnet.binancefuture.com"

# Phemex
PHEMEX_FUT_BASE_URL = "https://vapi.phemex.com"
PHEMEX_FUT_TEST_BASE_URL = "https://testnet-api.phemex.com"

# Kucoin
KUCOIN_FUT_BASE_URL = "https://api-futures.kucoin.com"
KUCOIN_FUT_TEST_BASE_URL = "https://api-sandbox-futures.kucoin.com"

# Bybit
BYBIT_FUT_BASE_URL = "https://api.bybit.com"
BYBIT_FUT_TEST_BASE_URL = "https://api-testnet.bybit.com"

# FTX
FTX_FUT_BASE_URL = "https://ftx.com/api"

# FTX-US
FTXUS_FUT_BASE_URL = "https://ftx.us/api"



from setuptools import setup, find_packages

VERSION = '1.0.10'
DESCRIPTION = "An algorithmic trading python package to save thousands of hours with the quickest crypto trading library so you can easily algo trade on any exchange & execute with unified function calls"
LONG_DESCRIPTION = "the onecall trading bot python library saves you thousands of hours to connect and trade with cryptocurrency exchanges. Taking out all of the hard work of connected to many exchanges and figuring out their documentation. with onecall you simply connect to one call and all of the functions for each exchange are the same so you save time and don't have to read new docs. onecall has a ton of custom algo trading functions built in saving you thousands of hours of time"

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="onecall",
    version=VERSION,
    author="Moon Dev",
    author_email="moondevonyt@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[
        'setuptools>=60.9.0',
        'requests>=2.18.4',
    ],

    keywords=['python', 'onecall', 'trading bot','quant', 'trading', 'trading algo', 'algorithmic trading', 'algo trading', 'crypto', 'exchange', 'automate trading', 'pinescript', 'mev', 'crypto exchange', 'binance', 'ftx', 'ethereum', 'solana', 'bitcoin', 'coinbase', 'phemex', 'bybit'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    include_package_data=True
)
